/*
** 
**   IRC - Internet Relay Chat, lib/gline.c
**   Copyright (C) 2001 Adam Loghry
**
**   Portions Copyright (C) 1999 Chris Behrens (last 2 functions)
**   Portions Copyright (C) 2001 "srvx development team" (buildglinelen && timechartype)
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**      
**   You should have received a copy of the GNU General Public License 
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include "ircd_includes.h"

RCSID("$Id: gline.c,v 1.1.1.1 2001/12/18 07:39:55 myr Exp $");

/*
 * still to do:
 * show_glines() should check for nick bans.. 
 * use queue'ing in gline deleting/expiring (but not adding)
 * 
 * possibles:
 * find nicks(/quote gline +lamer 1234567 :you suck, lamer)
 * reference id's on glines(so you can /quote gline -referenceid and not find a host) .. could suck or be really cool, but not in between
 * multiple glines in one command(/gline +lamer.com,lamer.net,lamer.org 1234456 :HAH! DIE!)
 *
 */

#ifdef GLINE_COUNT_CHECKING
int 	total_glines = 1;
#else
int	total_glines = 0;
#endif

int 	remote_gline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args);
int	check_usertotal(char *ban);
unsigned long buildglinelen(const unsigned char *interval);
static long timechartype(char type);

int m_gline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	/*
	 * /gline +host.com 1w :fuck off
	 * /gline +user@host.com 1w :fuck off
	 * /gline +nick!user@host.com 12 :fuck off
	 *
	 * /gline !+blah 1w :fuck off
	 * 
	 */

	char	*ban;
	char	*expire;
	char	*reason;	
	int	isforcing	= 0;
	char	*pm		= "+";
	int	hasnick		= 0;
	int	hasuser		= 0;

	if(IsServer(lptr))
		return remote_gline(lptr, pptr, prefix, args);
	else
	{
		ban 	= ircd_nextarg(&args, 0);
		expire  = ircd_nextarg(&args, 0);
		reason	= ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	}

	if(ban && IsAnOper(pptr))
	{
		if(*ban == '!')
		{
			isforcing = 1;
			ban++;
		}

		if(*ban == '+')
			pm = "+";
		else if (*ban == '-')
			pm = "-";
	}

	if(IsAnOper(pptr) && *pm == '+' && expire && reason)
	{
#if 0
		time_t  	timeout		= (atoi(expire) + Now);
#endif
		time_t		timeout		= buildglinelen(expire);
		int 		targets 	= 0;
		char		finalban[USERLEN+HOSTLEN+NICKLEN+2];
		IRCD_GLINE	*glptr;

		ban++;

		if(strchr(ban, '!'))
			hasnick = 1;
		if(strchr(ban, '@'))
			hasuser = 1;
		
		if(!hasnick && !hasuser)
		{
			/* "lamer.shellhost.com" */
			sprintf(finalban, "*@%s", ban);
		}
		else if(!hasnick && hasuser)
		{
			/* "lamer@lamer.shellhost.com" */
			strcpy(finalban, ban);
		}
		else if(hasnick && hasuser)
		{
			/* "lamer!lamer@lamer.shellhost.com" */
			strcpy(finalban, ban);
		}
		else
		{
			/* eh? invalid! */
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Invalid host to ban, G-Line rejected.", pptr->name);
			return 3;
		}
#if 0	/* done in buildglinelen() now */
		/* check expire time for sanity */
		if(timeout <= 0)
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Invalid expire time! (%s)",
							pptr->name, support_ctime(timeout));
			return 3;
		}

		if(timeout < Now) /* possible with a /quote gline +*!*@*.suck.com -91395913991391395193 :hi */
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Trying to gline in the past, eh? (%i)",
							pptr->name, support_ctime(timeout));
			return 3;
		}
#endif
		if(timeout == -1)
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Invalid expire time!", pptr->name);
			return 3;
		}

		/* check usercount, if its > 20 then drop it(unless they are forcing) */
		if(!isforcing)
			targets = check_usertotal(finalban);

		if(targets > 20)
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :G-Line rejected, affects too many(>20) local users; use !+%s to force it.",
					pptr->name, ban);
			return 3;
		}

		if(targets == Globalinfo->g_clients)
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :G-Line rejected, affects everyone on the network; use !+%s to force it.",
					pptr->name, ban);
			return 3;
		}

		/* and finally, check if its already banned */
		if((glptr = find_gline_bymask(finalban)) != NULL)
		{
			send_prefix_to_client(meptr, pptr, 
					"NOTICE %s :G-Line rejected, %s is already banned by %s!",
						pptr->name, finalban, glptr->ban);
			return 3;
		}
	
		/* send wallops to our guys */
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
					"%s[@%s] is%s adding G-Line on: %s (until: %s)",
						pptr->name, meptr->name, (isforcing) ? " forcibly" : "", finalban, 
								(timeout != 0) ? support_ctime(timeout) : "(permanent, does not expire)");
#if 0
		/* send wallops to other servers */
		send_to_servers(NULL, ":%s WALLOPS :%s[@%s] is%s adding G-Line on: %s (until: %s)",
					meptr->name, pptr->name, meptr->name, (isforcing) ? " forcibly" : "", finalban, support_ctime(timeout));
#endif
		/* send out new gline command (:server.name GLINE +*@banned.mask oper_nick expire_time total_glines :reason_for_ban) */
		send_to_servers(NULL, ":%s GLINE +%s %s %i %i :%s",
					meptr->name, finalban, pptr->name, timeout, total_glines, reason);

		/* add gline to our list */
		add_gline(finalban, timeout, pptr->name, reason);
		
		/* we're done, foo */
		return 3;
	}
	else if(IsAnOper(pptr) && *pm == '-' && ban)
	{
		char            finalban[USERLEN+HOSTLEN+NICKLEN+1];
		char		ubreason[KILLLEN] = ""; /* good enough length as any */
		IRCD_GLINE	*glptr = NULL;
		ban++;

		if(expire)
			sprintf(ubreason, "(reason: %s)", expire);
		if(reason)
			sprintf(ubreason, "(reason: %s %s)", expire, reason);

		if(strchr(ban, '!') && strchr(ban, '@'))		/* nick!user@host */
			strcpy(finalban, ban);
		else if(!strchr(ban, '!') && !strchr(ban, '@'))		/* host */
			sprintf(finalban, "*@%s", ban);
		else							/* probably user@host */
			strcpy(finalban, ban);

		if((glptr = find_gline_bymask(finalban)) == NULL)
		{
			if(isforcing)		/* using ! on a ungline will cause the gline to be propogated regardless of whether or not i have it */
			{
				send_to_servers(NULL, ":%s GLINE !-%s %s %i :%s",
							meptr->name, finalban, pptr->name, total_glines, ubreason);
				send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr, 
							"%s[@%s] is forcing removal of non-existant G-Line: %s %s",
								pptr->name, meptr->name, finalban, ubreason);
				send_to_servers(NULL, ":%s WALLOPS :%s[@%s] is forcing network-wide removal of non-existant G-Line: %s",
								meptr->name, pptr->name, meptr->name, finalban);
			}
			
			/* arr, no gline found. */
			send_prefix_to_client(meptr, pptr, "NOTICE %s :No G-Lines found matching %s! %s", pptr->name, finalban,
					isforcing ? "(propogating change to network)" : "");
			
			return 3;
		}

		/* we're sure there was a gline, so lets get going.. */
#ifdef GLINE_COUNT_CHECKING
		--total_glines;
#endif
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
					"%s[@%s] is removing G-Line: %s %s",
						pptr->name, meptr->name, glptr->ban, ubreason);
	
		send_to_servers(NULL, ":%s GLINE -%s %s %i :%s", 
					meptr->name, finalban, pptr->name, total_glines, ubreason);
#ifdef GLINE_COUNT_CHECKING
		if(total_glines < 0)
		{
			send_to_opers("ack!! total number of glines somehow got below zero!! (%i)", total_glines);
			total_glines = 0;
		}
#endif

		del_gline_from_gline_list(&Glines, glptr);
		MyFree(glptr);

		return 3;
	}
	else /* not +/- .. they're checking for the existence of a gline */
	{
		if(!ban)	/* if they just /gline, they are checking if THEY are glined. */
		{
			char tmp[NICKLEN+USERLEN+HOSTLEN+2];
			sprintf(tmp, "%s@%s", pptr->username, pptr->hostname);
			show_glines(pptr, tmp, 1);
		}
		else 
		{
			if(!strchr(ban, '@'))
			{
				send_prefix_to_client(meptr, pptr, "NOTICE %s :Incorrect host format (user@host)", pptr->name);
				return 3;
			}

			show_glines(pptr, ban, 0);
		}


		send_prefix_to_client(meptr, pptr,
		            IRCD_REPLY(RPL_ENDOFSTATS, pptr->name), 'G');

		return 3;
	}

	/* not reached */
	return 3;
}

int remote_gline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	/* we do all our own parsing here(instead of above, in m_gline())
	 *
	 * remote gline being added:		:server.network.net GLINE +*!*@*.banned.hostmask.com nick-of-oper expire-time total_number_of_glines :reason
	 * remote gline being removed:		:server.network.net GLINE -*!*@*.banned.hashmark.com nick-of-oper total_number_of_glines :reason
	 *
	 * the total number of glines is checked with each add/remove for purposes of keeping sync on the network;
	 * each server should have the same gline list. it isnt going to be strictly enforced though, but if another server has a diff
	 * number of glines then the ircd will scream about it to its opers... might be removed in the future if its not usefull, b ut
	 * seems like a decent enough behind-the-scenes idea for now.. */

	/* if the remote server sends out a gline with a total_number that is zero, the whole G-Line checking process is passed over.. 
	 * 		this is done so services dont have to track this kind of a #, although i dont see why they cant. */

	char	*ban		=	ircd_nextarg(&args, 0);
	char	*oper		=	ircd_nextarg(&args, 0);
	char	*number 	= 	ircd_nextarg(&args, 0);
#if 0
	char	*number2	=	ircd_nextarg(&args, 0);
	char	*reason		=	ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
#endif
	
	char	*pm		= 	NULL;
#ifdef GLINE_COUNT_CHECKING
	extern  int total_glines;
#endif
	IRCD_CLIENT *acptr	=	NULL;
	IRCD_GLINE  *glptr	= 	NULL;
	int	isforcing	= 	0;

	acptr = hash_find_client(oper);

	if(!is_ulined(prefix) && !is_ulined(pptr->servername) && !acptr)
		 return 3; /* dont accept fake oper nicks from non-ulined servers */

	/* locops with cangline can use gline, so this is disabled. */
#if 0
	if(!is_ulined(acptr->servername) && !IsAnOper(acptr))
		return 3;
#endif

	if(!ban || !oper || !number)
		return 3;

	if(*ban == '!')
	{
		isforcing = 1;
		ban++;
	}
	
	if(*ban == '+')
		pm = "+";
	else if(*ban == '-')
		pm = "-";
	else
		return 3; /* bad prefix on gline(not +/-) */

	ban++; 

	glptr = find_gline_bymask(ban);

	if(*pm == '+')
	{
		/* ban oper number 
		 * :prefix GLINE +ban oper number number2 :reason */
		char    *number2        =       ircd_nextarg(&args, 0);
		char    *reason         =       ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
		/* :prefix GLINE (+)host oper number number2 :reason */
		time_t 		expire = atoi(number);
#ifdef GLINE_COUNT_CHECKING
		int		rtg    = atoi(number2);
#endif
		if(!reason || !number2)
			return 3; /* arr */
		
		if(expire <= 0 || expire < Now)
			return 3;

		if(glptr && isforcing)	/* its glined, and the oper is forcing propogation.. so we propogate and DONT ADD */
		{
			send_to_servers(lptr, ":%s GLINE !+%s %s %s %s :%s",
							ban, oper, number, number2, reason);
			return 3;
		}
		else if (glptr)
		{
			if(CliGotEOB(lptr))
			{
				send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr, "Possible Desync: Recieved duplicate G-Line on %s(already banned by: %s) from %s",
							ban, glptr->ban, prefix);
				send_to_servers(NULL, ":%s WALLOPS :Possible Desync: Recieved duplicate G-Line on %s(Already banned by: %s) form %s",
							meptr->name, glptr->ban, prefix);
			}
			return 3;
		}

#ifdef GLINE_COUNT_CHECKING		
		if((rtg != 0) && (rtg != total_glines) && CliGotEOB(lptr))
		{
			send_to_opers("Possible Desync: %s has a different G-Line count than %s; %i(%s) %s %i(%s)",
					prefix, 
					meptr->name, 
					rtg, prefix, 
					(rtg > total_glines) ? ">" : "<", 
					total_glines, meptr->name);
#if 0
			if(!ImAHub)
				total_glines = number2;
			else
				/* ..? */ ;
#endif

		}

		/* okay, we've checked time, our gline count, their oper's authenticity.. we're set */
		++total_glines;
#endif
		if(CliGotEOB(pptr))
		{
			send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
							"%s[@%s] is%s adding G-Line on: %s (until: %s)",
								oper, prefix, 
								(isforcing) ? " forcibly" : "",
								ban, support_ctime(expire));
				
			send_to_servers(lptr, ":%s GLINE +%s %s %s %s :%s",
							 prefix, ban, oper, number, number2, reason);
		}
#ifdef DEBUG_MSGS
		send_to_opers(":%s GLINE +%s %s %s %s :%s",
						 meptr->name, ban, oper, number, number2, reason);
#endif
		add_gline(ban, expire, oper, reason);

		return 3;
	}
	else if(*pm == '-') /* it is */
	{
		char	*ubreason 	= ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
		/* :prefix GLINE (-)host oper number */
#ifdef GLINE_COUNT_CHECKING
		int	rtg = atol(number);
#endif
		if(!ubreason)
			strcpy(ubreason, ":");
		
		if(!glptr && isforcing)
		{
			send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
							"%s[@%s] is forcibly removing non-existant G-Line: %s %s",
								oper, prefix, ban, ubreason);
			send_to_servers(lptr, ":%s GLINE !-%s %s %s :%s",
							prefix, ban, oper, number, ubreason);
			return 3;
		}
		else if(!glptr)
		{
			send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr, "Possible Desync: Recieved request to remove G-Line on %s(which is not banned) from %s",
							ban, prefix);
			send_to_servers(NULL, ":%s WALLOPS :Possible Desync: Recieved request to remove G-Line on %s(which is not banned) from %s",
							meptr->name, ban, prefix);
			return 3;
		}
#ifdef GLINE_COUNT_CHECKING
		if((rtg != 0) && (rtg != total_glines))
		{
			send_to_opers("Possible Desync: %s has a different G-Line count than %s; %i(%s) %s %i(%s)",
					prefix, meptr->name, 
					rtg, prefix, 
					(rtg > total_glines) ? ">" : "<", 
					total_glines, meptr->name);
		}
#endif
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
					"%s[@%s] is removing G-Line: %s %s",
						oper, prefix, ban, ubreason);
#ifdef GLINE_COUNT_CHECKING
		send_to_servers(lptr, ":%s GLINE -%s %s %i %s",
						prefix, ban, oper, rtg, ubreason);
#else
		send_to_servers(lptr, ":%s GLINE -%s %s 0 %s",
						prefix, ban, oper, ubreason);
#endif
#ifdef DEBUG_MSG
		send_to_opers(":%s GLINE -%s %s %s",
						meptr->name, ban, oper, number);
#endif
#ifdef GLINE_COUNT_CHECKING
		--total_glines;
#endif
		del_gline_from_gline_list(&Glines, glptr);
		MyFree(glptr);

		return 3;
	}
		
	return 3;
}
/* ** for local && remote clients ** */
int add_gline(char *ban, time_t expire, char *oper, char *reason)
{
		/* all sanity checking is performed before arriving here */
		IRCD_GLINE *newgline;
#ifdef GLINE_COUNT_CHECKING
		extern int total_glines;
#endif

		if(strlen(reason) > KILLLEN)
			reason[KILLLEN] = '\0'; /* truncate them crazy gline msgs */
		
		newgline = (IRCD_GLINE *)MyMalloc(sizeof(IRCD_GLINE));

		newgline->ban		= MyStrdup(ban);
		newgline->expire 	= expire;
		newgline->hasnick	= (strchr(ban, '!')) ? 1 : 0;
		newgline->oper 		= MyStrdup(oper);
		newgline->reason	= MyStrdup(reason);
		if(strchr(newgline->ban, '!'))
			newgline->hasnick = 1;
		else
			newgline->hasnick = 0;

		add_gline_to_gline_list(&Glines, newgline);

		killall_by_match(ban, reason, oper, expire, meptr->name);
#ifdef GLINE_COUNT_CHECKING
		++total_glines;
#endif
		return 1;
}

IRCD_GLINE *find_gline_bymask(char *ban)
{
	IRCD_GLINE *gptr = Glines;
#ifdef GLINE_COUNT_CHECKING
	extern int total_glines;

	if(!total_glines)
		return NULL;
#endif
	
	for(;gptr;gptr=gptr->next)
	{
		if(gptr->ban == NULL)
			break; /* ::sigh:: */

		if(!match(gptr->ban, ban))
			return gptr;
	}

	return NULL;
}

IRCD_GLINE *check_newuser_glined(IRCD_CLIENT *cptr)
{
	char 		usermask[NICKLEN+USERLEN+HOSTLEN+1];
	char		usermask2[USERLEN+HOSTLEN+1];
	IRCD_GLINE *gptr = Glines;


	sprintf(usermask, "%s!%s@%s",
			cptr->name, cptr->username, cptr->hostname);
	sprintf(usermask2, "%s@%s", cptr->username, cptr->hostname);

	if(!gptr)
		return NULL;

	for(;gptr;gptr=gptr->next)
	{
		if(gptr->expire < Now)
		{
			expire_gline(gptr);
			continue;
		}
		if(gptr->hasnick)
		{
			if(!match(gptr->ban, usermask2))
				return gptr;
		}
		else
			if(!match(gptr->ban, usermask))
				return gptr;
	}

	return NULL;
}

void show_glines(IRCD_CLIENT *cptr, char *ban, int ishim)
{
#ifdef GLINE_COUNT_CHECKING
	extern int total_glines;
#endif
	IRCD_GLINE *gptr = Glines;
	int x = 0;
	int sh = (int)vars_value_get_bynum(IRCD_VAR_MAX_GLINESEARCH);

	if(!ban)
	{
		for(;gptr;gptr=gptr->next)
		{
			if(gptr->expire < Now)
			{
				expire_gline(gptr);
				continue;
			}
								/* %c %s@%s %s :%s (%s) */
			send_prefix_to_client(meptr, cptr, IRCD_REPLY(RPL_STATSGLINE, cptr->name),
									'G', gptr->ban, support_ctime(gptr->expire), gptr->reason, gptr->oper);
							
			x++;
			if (x > sh)
			{
			      	send_prefix_to_client(meptr, cptr, "NOTICE %s :Too many results! Specify a less-broad request next time [>%i]",
							cptr->name, sh);
				break;
			}
		}
	}
	else
	{
		for(;gptr;gptr=gptr->next)
		{
			if(gptr->expire < Now)
			{
				expire_gline(gptr);
				continue;
			}
#ifdef DEBUG_MSGS
			send_to_opers("%s cmp %s %c %c ", gptr->ban, ban, 
					(!match(gptr->ban, ban)) ? 'y' : 'n',
					(!match(ban, gptr->ban)) ? 'y' : 'n');
#endif

			/* if(!match(gptr->ban, ban)) */
			if(!match(ban, gptr->ban))
			{
				send_prefix_to_client(meptr, cptr, IRCD_REPLY(RPL_STATSGLINE, cptr->name),
								'G', gptr->ban, support_ctime(gptr->expire), gptr->reason, gptr->oper);
				x++;
			}

			if(x > sh)
			{
		 		send_prefix_to_client(meptr, cptr, "NOTICE %s :Too many results! Specify a less-broad request next time [>%i]",
							cptr->name, sh);
                                break;
			}

		}

		if(!x && ishim)
			send_prefix_to_client(meptr, cptr, "NOTICE %s :You are not G-Lined.", cptr->name);
	}
#ifdef GLINE_COUNT_CHECKING
	if(!ishim)
	{
		if(x)
			send_prefix_to_client(meptr, cptr, "NOTICE %s :%i/%i G-Lines Found.",
					cptr->name, x, total_glines-1);
		else
			send_prefix_to_client(meptr, cptr, "NOTICE %s :No G-Lines found.", cptr->name);
	}
#endif

	send_to_flag(CLIENT_NFLAGS_STATSNOTICES, NULL,
				"G-Line search performed by: %s on: %s (%i/%i)",
					cptr->name, (ban) ? ban : "*@*", x, total_glines);
}

void expire_gline(IRCD_GLINE *gptr)
{
	send_to_flag(CLIENT_NFLAGS_GLINEEXPIRES, NULL, "Removing G-Line: %s (expired at: %s)", gptr->ban, support_ctime(gptr->expire));
	del_gline_from_gline_list(&Glines, gptr);
	MyFree(gptr);

	return;
}

int check_usertotal(char *ban)
{
	char 		uh[NICKLEN+USERLEN+HOSTLEN+1];
	char		saneban[NICKLEN+USERLEN+HOSTLEN+1];
	IRCD_CLIENT	*cptr;
	int		i = 0;

	if(!strchr(ban, '!') && !strchr(ban, '@'))			 /* ban is in full nick!user@host form */
		sprintf(saneban, "*!*@%s", ban);
	else if(!strchr(ban, '!') && strchr(ban, '@'))			 /* ban is in user@host form */
		sprintf(saneban, "*!%s", ban);
	else								 /* ban is just a host */
		sprintf(saneban, "%s", ban);

	for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
	{
		sprintf(uh, "%s!%s@%s",
				cptr->name, cptr->username, cptr->hostname);
		if((!match(saneban, uh)) && !(IsServer(cptr) || IsService(cptr) || IsAnOper(cptr) || is_ulined(cptr->servername)))
			i++;
	}

	return i;
}

void killall_by_match(char *ban, char *why, char *oper, int timeout, char *servername)
{
	char		usermask[NICKLEN+USERLEN+HOSTLEN+1];
	char		reason[KILLLEN];
	IRCD_CLIENT *cptr;

	for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
	{
		sprintf(usermask, "%s!%s@%s", cptr->name, cptr->username, cptr->hostname);
	
		if (!match(ban, usermask))
		{
			if(IsServer(cptr))
				continue; 	/* heh i had to find this out the hard way to remember to check that.. doh -myriad */
			if(IsAnOper(cptr) || IsService(cptr) || is_ulined(cptr->servername))
			{
				send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
					"%s[@%s] matches G-Line by %s on %s but will NOT be killed! (%s)",
						cptr->name, cptr->servername, oper, ban,
						(IsAnOper(cptr)) ? "oper" : (IsService(cptr)) ? "service" : "ulined");
				send_to_servers(NULL, ":%s WALLOPS :%s[@%s] matches G-Line by %s on %s but will NOT be killed! (%s)",
						meptr->name, cptr->name, cptr->servername, oper, ban,
						(IsAnOper(cptr)) ? "oper" : (IsService(cptr)) ? "service" : "ulined");
				send_prefix_to_client(meptr, cptr, "NOTICE %s :%s[@%s] added a G-Line that affects you: %s",
						cptr->name, oper, servername, ban);
				continue;
			}
			else
			{
				send_prefix_to_client(meptr, cptr, "NOTICE %s :*** You are no longer welcome on this network.",
							cptr->name);
				if(vars_value_get_bynum(IRCD_VAR_SHOW_GLINE_MSGS))
				{
					send_prefix_to_client(meptr, cptr, "NOTICE %s :*** You have been G-Lined until %s by %s[@%s].",
							cptr->name, support_ctime(timeout), oper, servername);
					send_prefix_to_client(meptr, cptr, "NOTICE %s :*** Reason: %s", cptr->name, why);
				}
				cptr->c_flags |= CLIENT_FLAGS_KILLED;
				sprintf(reason, "G-Lined: %s (%s)", why, support_ctime(timeout));
				client_del(meptr, cptr, reason);
				continue;
			}
		}
	}
	return;
}

/* these two functions(timechartype & buildglinelen) are taken from srvx's code; it was well done there and i 
 * saw no need to reinvent the wheel. this whole thing makes doing glines a hell of a lot
 * easier.. -myr */
static long timechartype(char type)
{
    switch (type) {
		case 'y': return 365*24*60*60;
	        case 'M': return 31*24*60*60;
	    	case 'w': return 7*24*60*60;
	        case 'd': return 24*60*60;
	        case 'h': return 60*60;
	        case 'm': return 60;
		case 's': return 1;
	        default: return 0;
	}
}

unsigned long buildglinelen(const unsigned char *interval)
{
	unsigned long seconds = 0;
	int partial = 0;
	char c;

	if(!strncasecmp(interval, "!P", 2))
		return 0; /* !P = perm. */

	if(!strchr(interval, 'y') && !strchr(interval, 'M') && !strchr(interval, 'w') && !strchr(interval, 'd') && 
			!strchr(interval, 'h') && !strchr(interval, 'm') && !strchr(interval, 's'))
		return atoi(interval) + Now; /* no time char stuff */

	while ((c = *interval++)) {
	       if (isdigit((int)c)) {
		            partial = partial*10 + c - '0';
		} else {
		            seconds += timechartype(c) * partial;
	        	    partial = 0;
		}
	}

	return seconds + partial + Now;
}

/* hash.c */
void add_gline_to_gline_list(IRCD_GLINE **all, IRCD_GLINE *newgline)
{
	newgline->last = NULL;
	if ((newgline->next = *all) != NULL)
			newgline->next->last = newgline;
	*all = newgline;
}

void del_gline_from_gline_list(IRCD_GLINE **all, IRCD_GLINE *delgline)
{
	if (delgline->last)
		delgline->last->next = delgline->next;
	else
		*all = delgline->next;
	if (delgline->next)
		delgline->next->last = delgline->last;
}
